#!/usr/bin/env perl
#
# gpxAdjust.pl, created 20190612 - 10:54:10 by ed
#
# Skeleton created by crtprlskl.pl.
#
# An emax tool.
#

use strict;
use warnings;

use Getopt::Long qw(:config no_ignore_case);
use POSIX;
use Time::HiRes qw(gettimeofday);
use Time::Local;
use Data::Dumper;

$Data::Dumper::Sortkeys  = 1;
$Data::Dumper::Indent    = 1;
$Data::Dumper::Quotekeys = 0;

my ($Help,$Debug,$Offset,$Unit,$Split,$Preserve) = (undef,'');

# remember epoch rollover: Tuesday, January 19th 2038, 03:14:07 UTC  ...
my $LastTS = 2147483647; # max epoch
my $Seconds = 0;
my $Now = time();
my $_1024Weeks = 1024*7*24*3600; # seconds of 1024 weeks
my $IsNewDay;
my $CurrentDay;

# gps start is January 6th, 1980, rollover events occur every 1024 week, 
#     e.g. August 22nd, 00:00, 1999, April 7th, 2019, 00:00 etc.
# (all dates are UTC)
# NOTE: This timestamps are simplified to midnight, in reality they occur e.g. April 6th, 2019, 23:59:42
my $FirstRO = timelocal(0,0,0,22,(8-1),1999);
my $LastRO;                     # last rollover before NOW

# to trace returns from subroutines, just add 'my $DbgLvl = new DbgLvl' at any soubroutine
{ package DbgLvl; sub new { my ($class, $self) = (shift, {});  bless $self; return $self; }
  sub DESTROY() { ::DEBUG('<-returning');} }

sub Usage() {
    print <<EOF;

    gpxAdjust.pl - adjust gpx timestamps in gpx-file. An emax-taool. 

    Usage:   gpxAdjust.pl [-o offset] [-p] [-s] [-h] [-D string] < input > output
     
    Mandatory args:
     gpxfile        the name of the gpxfile to be processed.
                   
    Optional args:
     -o offset       sets the time-offset by which timestamp shall be 
                     corrected. 'offset' can have the format
                       [-]hh:mm:ss where the leading minus-sign is optional.
                       [-]nw       where 'n' is an integer value and 'w' 
                                   defines this number to mean weeks.
                       [-]nd       where 'n' is an integer value and 'd' 
                                   defines this number to mean days.

     -p              preserves date values after last rollover
                     If set, dates after the last rollover stay unchanged
                     
     -s              Splits the tracks into tracksegments upon day-change.
     -h              prints this screen and exits
     -D {string}     prints debugging informations according to 'string'.
                     'string' may contain any of the following charaters:                        
                        A    :    dump \@ARGV
                        a    :    dump all
                        o    :    dump options
                        R    :    dump rollover values
                        O    :    dump calculated offset seconds
    Examples:

     gpxAdjust -o 1024w < holux.gpx > corrected.gpx
       Reads 'holux.gpx' and adds 1024 weeks to each timestamp. 
       The output is written to 'corrected.gpx'.

     gpxAdjust -o -02:00:00 < germay.gpx > utc.gpx       
       Reads 'germany.gpx' and subtracts 2 hours from each timestamp. 
       The output is written to 'utc.gpx'.

     NOTE: one of '-o' or '-s' options which must be specified.
EOF

1;
} # Usage

sub GetISODateStamp {
    my ( $eseconds, $msec ) = gettimeofday();
    my ( $second, $minute, $hour, $dayOfMonth, $month, $yearOffset, $dayOfWeek, $dayOfYear, $daylightSavings ) = localtime(time);
    return sprintf "%02d-%02d-%02d %02d:%02d:%02d.%03d", ( $yearOffset - 100 ), ( $month + 1 ), $dayOfMonth, $hour, $minute, $second, $msec / 1000;
} # GetISODateStamp

sub SetDbg($) {
    my ($set) = @_;
    my $caller_pkg = caller;
    no strict qw(refs);

    defined *{"$caller_pkg\::DEBUG"} and undef *{"$caller_pkg\::DEBUG"};
    if (!$set) {
        *{"$caller_pkg\::DEBUG"} = sub { return 1; };
    } else {
        *{"$caller_pkg\::DEBUG"} = sub { 
            my $dbgmsg = shift;
            # get nesting level
            my $level = 0;
            1  while caller(++$level);
            --$level;            # $level -= $baselevel; alternative for packagaes
            # get stack information
            my $line = sprintf('%04d',(caller 0)[2]); 
            my $subr = (caller 1)[3];
            !defined $subr and $subr = '';
            my $adjust = 0;
            if ($subr =~ /DbgLvl::DESTROY/) {
                $subr = (caller 3)[3].' <<';
                $level -= 2;
            }
            $subr =~ s/^main:://g;
            ! length $subr || $subr eq '(eval)' and $subr = '     -main-     ';
            if (!defined $dbgmsg or length $dbgmsg eq 0) { 
                my $filler = 30-$level;
                $dbgmsg = sprintf ("SUB %$filler.${filler}s> $subr", "=============================");
            } 

            while (length $subr > 16 && $subr =~ s/([^aeiou]+.*)[aeiou]/$1/g) {};
            $subr = sprintf("%-16.16s",$subr);
            my $indent = '.' x $level;
            printf STDERR ("%s %s:%s:%s\n", 
                           GetISODateStamp,
                           $line,$subr,$indent.$dbgmsg);        
            return 1;
        };
    }
} # setDbg

sub getLastRollover() {
    # get the last rollover date before (Now+1)
    my $lastRO = $FirstRO;    
    while ($lastRO <= $Now) {
        $lastRO += $_1024Weeks;        
    } # while lastRO < Now

    $lastRO -= $_1024Weeks;
    return $lastRO;
} # getLastRollover

sub replaceTS() {
    # replace an incoming '<time>1999-09-01T10:37:36Z</time>' with the corrected timestamp
    # NOTE:    only the date values are passed as $1..$6    
    # CAUTION: this function alters global variables    
    my ($tp, $yyyy,$mm,$dd,$hh,$mi,$ss) = @_;
    my $gpxtime = timelocal($ss,$mi,$hh,$dd,$mm,$yyyy);
    my $newtime = ($Preserve && $gpxtime > $LastRO) ? $gpxtime : $gpxtime + $Seconds;
    my ($newSS,$newMI,$newHH,$newDD,$newMM,$newYYYY) = localtime($newtime);
    $newYYYY += 1900;

    if ($CurrentDay) {
        my ($curSS,$curMI,$curHH,$curDD,$curMM,$curYYYY) = localtime($CurrentDay);
        $IsNewDay = (($newtime - (24*3600)) > $CurrentDay || $newDD > $curDD);
    }
    
    $CurrentDay = $newtime;
    
    return sprintf "<time>%04d-%02d-%02dT%02d:%02d:%02dZ<\/time>", $newYYYY,$newMM,$newDD,$newHH,$newMI,$newSS;
} # replaceTS

sub processTP() {
    my $tp = shift;
    $tp =~  s/<time>(\d\d\d\d)-(\d\d)-(\d\d)T(\d\d):(\d\d):(\d\d)Z<\/time>/&replaceTS($tp, $1, $2, $3, $4, $5, $6)/mge;    
    return $tp;
} # processTP

sub splitTrack() {    
    my ( $ss, $mi, $hh, $dd, $mm, $year ) = localtime($CurrentDay);

    printf "
    </trkseg>
  </trk>
  <trk>
    <name>Track-%04d%02d%02d_%02d%02d%02d</name>
    <desc>Track %04d.%02d.%02d %02d:%02d:%02d</desc>
  <trkseg>", 
    $year+1900, $mm, $dd, ,$hh, $mi, $ss,
    $year+1900, $mm, $dd, ,$hh, $mi, $ss;
} # splitTrack

#========================================
# main
#========================================

SetDbg(1); # activate debugging
#SetDbg(0); # deactivate debugging

GetOptions  ('help|h|?'      => \$Help, 
             'debug|D=s'     => \$Debug,
             'offset|o=s'    => \$Offset,
             'preserve|p'    => \$Preserve,
             'split|s'       => \$Split             
    );

if ($Debug =~ /a|o/) {    
    DEBUG("Options:"); 
    DEBUG(" help|h|?              : [".($Help     // '-undef-').']');
    DEBUG(" offset|o              : [".($Offset   // '-undef-').']');
    DEBUG(" preserve|p            : [".($Preserve // '-undef-').']');
    DEBUG(" split|s               : [".($Split    // '-undef-').']');
    DEBUG(" debug|D=s             : [".($Debug    // '-undef-').']');
} # Debug options

if ($Debug =~ /a|A/) {
    my $a = 0;
    foreach (@ARGV) {
        DEBUG("\$ARGV[$a] = [$_]");
        ++$a;
    } # foreach ARGV
} # Debug args

$Help and  &Usage() and exit;
!$Offset && !$Split and die "\nNo offset, no split: nothing to do. Try '-h'.\n\n";

if ($Offset) {
    # [-]hh:mm:ss where the leading minus-sign is optional.
    # [-]nw       where 'n' is an integer value and 'w' 
    #             defines this number to mean weeks.
    # [-]nd       where 'n' is an integer value and 'd' 
    #             defines this number to mean days.

    my $offs = $Offset;
    my $factor = ($offs =~ s/^-//g) ? -1 : 1;

    if ($offs =~ /^(\d+):(\d+):(\d+)$/) {
        $Seconds = $1*3600 + $2*60 + $3;
    } elsif ($offs =~ /^(\d+)w$/ ) {
        $Seconds = $1 * 7 * 24*3600;
    } elsif ($offs =~ /^(\d+)d$/ ) {
        $Seconds = $1 * 24*3600;
    } else {
        die "\n\nInvalid offset format.\n\n";
    }

    $Debug =~ /a|O/ and DEBUG("Offset factor:  [$factor]");
    $Seconds = $Seconds * $factor;
    $Debug =~ /a|O/ and DEBUG("Offset seconds:  [$Seconds]");
} # if Offset

$LastRO = &getLastRollover();

if ($Debug =~ /a|R/) {
    DEBUG("FirstRO:  [$FirstRO], [".(localtime $FirstRO)."]");
    DEBUG("LastRO:  [$LastRO], [".(localtime $LastRO)."]");
    DEBUG("Now:     [$Now], [".(localtime $Now)."]");
}

$/ = undef;  # unset input separator

while (<STDIN>) {
    # NOTE: STDIN is read as one long strea
    while (length) {
        my $isNewTrack = 0;        
        if (!/<trkpt/) {
            print "$_";                                            # no more trackpoint, flush
            last;
        }

        if (s/^(.*?)(?=<trkpt)//ms) {
            my $fill = "$1";                                       # print anything up to first trackpoint
            print $fill;
            $fill =~ /<trk>/ and $isNewTrack = 1;
        }
        
        while (/^\s*<trkpt/ && s/^(\s*<trkpt .+?)(?=<\/trkpt>)//ms) {        
            my $trackPoint = $1;                                   # we have a complete trackpoint
            s/^(\s*<\/trkpt>)//ms and  $trackPoint .= $1;          # consume pending '</trkpt>'
            $trackPoint = &processTP($trackPoint);
            !$isNewTrack && $IsNewDay && $Split and &splitTrack(); # if still in same track then split track if necessary
            print $trackPoint;
        }
    }
} # while stdin

print STDERR "\ndone.\n\n";
